#define PORTA_AUX_IO // required to enable external I/O bus

/* LCD control signals */
#define	LCD_A0	0


void MsDelay ( int iDelay )
{
	unsigned long ul0;
	ul0 = MS_TIMER;		// get current timer value
	while ( MS_TIMER < ul0 + (unsigned long) iDelay );
}


/* Low level functions */

void LCD_Write(unsigned char byte);
void LCD_WritePixel(unsigned long rgb);
void LCD_WriteArea(unsigned long rgb,unsigned long area);
void LCD_DumpArea(unsigned long img,unsigned long area);

#asm root
;function requires one parameter:
;@sp+2= data to write
;
LCD_Write::

	;ld hl,(sp+2)			; get data value (LSB)
	ld a,l
   ioe ld(0xB000),A
	ret

LCD_WritePixel::
	;ld bcde,(sp+2)			; get data value
	ld a,c
   ioe ld(0xB000),A
	ld a,d
   ioe ld(0xB000),A
	ld a,e
   ioe ld(0xB000),A
   ret

LCD_WriteArea::
	;ld bcde,(sp+2)			; get data value
   ex jkhl,bcde
	ld bcde,(sp+6)			; get count
   ex de,hl
   ex bc,hl
   ex de,hl
   ld		a,c
	or		b
	jr		nz,.l2
.l1:
	dec	de
.l2:
   ex jkhl,bcde
	ld a,c
   ioe ld(0xB000),A
	ld a,d
   ioe ld(0xB000),A
	ld a,e
   ioe ld(0xB000),A
   ex jkhl,bcde
   dwjnz .l2
   ld a,d
   or e
   jr nz,.l1
   ret

LCD_DumpArea::
	;ld bcde,(sp+2)			; get bitmap pointer
   ex jkhl,bcde
	ld bcde,(sp+6)				; get count
   ex de,hl
   ex bc,hl
   ex de,hl
   ld		a,c
	or		b
	jr		nz,.l4
.l3:
	dec	de
.l4:
   ex jkhl,bcde
   ld px,bcde
   ld bcde,(px+0)
	ld a,c
   ioe ld(0xB000),A
	ld a,d
   ioe ld(0xB000),A
	ld a,e
   ioe ld(0xB000),A
   ld bcde,px
   ex jkhl,bcde
   ld py,bcde
   ld bcde,3
   add jkhl,bcde
   ld bcde,py
   dwjnz .l4
   ld a,d
   or e
   jr nz,.l3
   ret

#endasm



void LCD_WriteCmd(unsigned char cmd)
{
	BitWrPortI ( PEDR, &PEDRShadow, 0,LCD_A0 );	// Lower A0 (Cmd)
	LCD_Write(cmd);
	BitWrPortI ( PEDR, &PEDRShadow, 1,LCD_A0 );	// Rise A0 (Data)
}

void LCD_WriteStrCmd(const unsigned char *cmd,int len)
{
	LCD_WriteCmd(*cmd++);
	while(len--) {
		LCD_Write(*cmd++);
	}
}

void LCD_setwindow (unsigned int x0,unsigned int x1,unsigned int y0,unsigned int y1 );


void LCD_init ()
{
const static unsigned char init_string1[]={
0xE0,0x01		// START PLL
};
const static unsigned char init_string2[]={
0xE0,0x03		// LOCK PLL
};

const static unsigned char init_string3[]={
0xB0,				//SET LCD MODE  SET TFT 18Bits MODE
0x0C,0x80,		//SET TFT MODE & hsync+Vsync+DEN MODE
0x02,0x7F,		//SET horizontal size=640-1
0x01,0xDF,		//SET vertical size=480-1
0x00 				//SET even/odd line RGB seq.=RGB
};

const static unsigned char init_string4[]={
0xF0,0x00 		//SET pixel data I/F format=8bit
};

const static unsigned char init_string5[]={
0x3A,0x60	   // SET R G B format = 6 6 6
};

const static unsigned char init_string6[]={
0xE6,0x02,0xFF,0xFF 		//SET PCLK freq=4.94MHz  ; pixel clock frequency
};

const static unsigned char init_string7[]={
0xB4,				//SET HBP,
0x02,0xF8,		//SET HSYNC Total=760
0x00,0x44,		//SET HBP 68
0x0F,				//SET VBP 16=15+1
0x00,0x00,		//SET Hsync pulse start position
0x00				//SET Hsync pulse subpixel start position
};

const static unsigned char init_string8[]={
0xB6,				//SET VBP,
0x01,0xF8,		//SET Vsync total
0x00,0x13,		//SET VBP=19
0x07,				//SET Vsync pulse 8=7+1
0x00,0x00		//SET Vsync pulse start position
};

// Use Port E bit 5 for LCD Chip Select with 3 wait-states
#define LCDSTROBE			0x20
#define LCDCSREGISTER	IB5CR
#define LCDCSSHADOW		IB5CRShadow
#define LCDCSCONFIG		0xC8

	WrPortI ( PEDR,&PEDRShadow,'\B01100111' );		// Make sure CS,RD,WR are HIGH
	WrPortI ( PEDDR,&PEDDRShadow,'\B01100111' );	// PE0,1,2,5,6 = output
//	WrPortI ( SPCR, &SPCRShadow, 0x84 );			// PA0-7 = Outputs
	// Initialize Port E bit to be a normal I/O pin
	WrPortI(PEFR, &PEFRShadow, (PEFRShadow|LCDSTROBE));
	// Initialize Port E bit to be an output pin
	WrPortI(PEDDR, &PEDDRShadow, (PEDDRShadow|LCDSTROBE));
	// Initialize Port E bit to be a chip select.
	WrPortI(LCDCSREGISTER, &LCDCSSHADOW, LCDCSCONFIG);
	// Set Port E bit to be clocked by PCLK/2
	WrPortI(PECR, &PECRShadow, (PECRShadow & ~0xFF));

	MsDelay ( 1000 );
	BitWrPortI ( PEDR, &PEDRShadow, 0,6 );	// Lower RST
	MsDelay ( 10 );
	BitWrPortI ( PEDR, &PEDRShadow, 1,6 );	// Raise RST
	MsDelay ( 100 );

	LCD_WriteCmd ( 0x01 );		// Software reset
	LCD_WriteCmd ( 0x01 );		// Software reset
	LCD_WriteCmd ( 0x01 );		// Software reset
	MsDelay ( 100 );
	LCD_WriteStrCmd ( init_string1,sizeof(init_string1) );
	LCD_WriteStrCmd ( init_string2,sizeof(init_string2) );
	LCD_WriteStrCmd ( init_string3,sizeof(init_string3) );
	LCD_WriteStrCmd ( init_string4,sizeof(init_string4) );
	LCD_WriteStrCmd ( init_string5,sizeof(init_string5) );
	LCD_WriteStrCmd ( init_string6,sizeof(init_string6) );
	LCD_WriteStrCmd ( init_string7,sizeof(init_string7) );
	LCD_WriteStrCmd ( init_string8,sizeof(init_string8) );
	LCD_setwindow ( 0, 639, 0, 479);
	LCD_WriteCmd ( 0x29 );		// Display ON
}

/* High level functions */
void LCD_setwindow ( unsigned int x0, unsigned int x1, unsigned int y0, unsigned int y1 )
{
	LCD_WriteCmd(0x2A);									// SET column address
	LCD_Write(((unsigned)x0>>8)&0xFF);							// HI byte
	LCD_Write(x0&0xFF); 									// LO byte
	LCD_Write(((unsigned)x1>>8)&0xFF);							// HI byte
	LCD_Write(x1&0xFF); 									// LO byte
	LCD_WriteCmd(0x2B);									// SET page address
	LCD_Write(((unsigned)y0>>8)&0xFF);							// HI byte
	LCD_Write(y0&0xFF); 									// LO byte
	LCD_Write(((unsigned)y1>>8)&0xFF);							// HI byte
	LCD_Write(y1&0xFF); 									// LO byte
}


void LCD_fill(unsigned int x, unsigned int y, unsigned long color)
{
	LCD_WriteCmd(0x2C);
  	LCD_WriteArea(color,(unsigned long)x * y);	// while(x*y--) LCD_WritePixel(color)
}

#define LCD_clear() 	LCD_setwindow(0,639,0,479);LCD_fill(640,480,0)


void LCD_icon(unsigned int x, unsigned int y, unsigned int wx, unsigned int wy, unsigned long img)
{
	LCD_setwindow ( x, wx+x-1, y, wy+y-1);
	LCD_WriteCmd(0x2C);
  	LCD_DumpArea(img+4,(unsigned long)wx * wy);	// while(wx*wy--) LCD_WritePixel(*img++)
}

QUADS()
{
unsigned int i,j;

	LCD_setwindow ( 0, 319, 0, 239);
	LCD_WriteCmd(0x2C);
  	LCD_WriteArea(0xFF0000,320UL*240);	//
	LCD_setwindow ( 320, 639, 0, 239);
	LCD_WriteCmd(0x2C);
  	LCD_WriteArea(0x0000FF,320UL*240);	//
	LCD_setwindow ( 0, 319, 240, 479);
	LCD_WriteCmd(0x2C);
  	LCD_WriteArea(0xFFFF00,320UL*240);	//
	LCD_setwindow ( 320, 639, 240, 479);
	LCD_WriteCmd(0x2C);
  	LCD_WriteArea(0x00FF00,320UL*240);	//
	LCD_setwindow ( 0, 639, 0, 479);
}


LCD_plot(unsigned int x,unsigned int y,unsigned long color)
{
	LCD_setwindow (x,x,y,y);
	LCD_WriteCmd(0x2C);
	LCD_WritePixel(color);
}

// Bresenham Line Algorithm

void LCD_line(int x1,int y1,int x2,int y2, unsigned long color)
{
int  x,y,dx,dy,incx,incy,balance;

        if (x2 >= x1){
                dx = x2 - x1;
                incx = 1;
        }
        else {
                dx = x1 - x2;
                incx = -1;
        }

        if (y2 >= y1) {
                dy = y2 - y1;
                incy = 1;
        }
        else {
                dy = y1 - y2;
                incy = -1;
        }

        x = x1;
        y = y1;

        if (dx >= dy) {
                dy <<= 1;
                balance = dy - dx;
                dx <<= 1;

                while (x != x2) {
                        LCD_plot(x, y,color);
                        if (balance >= 0) {
                                y += incy;
                                balance -= dx;
                        }
                        balance += dy;
                        x += incx;
                }
                LCD_plot(x, y,color);
        }
        else {
                dx <<= 1;
                balance = dx - dy;
                dy <<= 1;

                while (y != y2) {
                        LCD_plot(x, y,color);
                        if (balance >= 0) {
                                x += incx;
                                balance -= dy;
                        }
                        balance += dx;
                        y += incy;
                }
                LCD_plot(x, y,color);
        }
}


#define LCD_rectangle(x1,y1,x2,y2,c) \
{ \
	LCD_line(x1,y1,x2,y1,c); \
	LCD_line(x2,y1,x2,y2,c); \
	LCD_line(x2,y2,x1,y2,c); \
	LCD_line(x1,y2,x1,y1,c); \
}


/* simple 3D function plotting */

void paraboloide(int PERSP,float SCALE,int TRAZOSX,int TRAZOSY,float X0,float X1,float Y0,float Y1,unsigned long pcolor)
{
auto int xx,yy;
auto float x,y,z;
auto float persp,xt,zt;
auto float h1,h2,h3,h4,cs,si;

	persp=PERSP*0.01745329;			// convert to radians
	cs=cos(persp);
	si=sin(persp);
	h1=(X1-X0)/159;
	h2=(Y1-Y0)/(float)(TRAZOSX-1);
	h3=(X1-X0)/(float)(TRAZOSY-1);
	h4=(Y1-Y0)/159;

	for(y=Y0;y<=Y1;y+=h2)
		for(x=X0;x<=X1;x+=h1){
			z=x*x/4+y*y/9-15;
			xt=x-y*cs;
			zt=z-y*si;
			xx=320+(int)(SCALE*xt);
			yy=240-(int)(SCALE*zt);
			LCD_plot(xx,yy,pcolor);
		}
	for(x=X0;x<=X1;x+=h3)
		for(y=Y0;y<=Y1;y+=h4) {
			z=x*x/4+y*y/9-15;
			xt=x-y*cs;
			zt=z-y*si;
			xx=320+(int)(SCALE*xt);
			yy=240-(int)(SCALE*zt);
			LCD_plot(xx,yy,pcolor);
		}
}

void elipsoide(int PERSP,float SCALE,int TRAZOSX,int TRAZOSY,float X0,float X1,float Y0,float Y1,unsigned long pcolor)
{
auto int xx,yy;
auto float x,y,z;
auto float persp,xt,yt,zt;
auto float h1,h2,h3,h4,cs,si;

	persp=PERSP*0.01745329;			// convert to radians
	cs=cos(persp);
	si=sin(persp);
	h1=(X1-X0)/159;
	h2=(Y1-Y0)/(float)(TRAZOSX-1);
	h3=(X1-X0)/(float)(TRAZOSY-1);
	h4=(Y1-Y0)/159;

	for(y=Y0;y<=Y1;y+=h2)
		for(x=X0;x<=X1;x+=h1){
			xt=cos(x)*cos(y);
			yt=sin(x)*cos(y);
			zt=sin(y);
			xt-=yt*cs;
			zt-=yt*si;
			xx=320+(int)(SCALE*xt);
			yy=240-(int)(SCALE*zt);
			LCD_plot(xx,yy,pcolor);
		}
	for(x=X0;x<=X1;x+=h3)
		for(y=Y0;y<=Y1;y+=h4) {
			xt=cos(x)*cos(y);
			yt=sin(x)*cos(y);
			zt=sin(y);
			xt-=yt*cs;
			zt-=yt*si;
			xx=320+(int)(SCALE*xt);
			yy=240-(int)(SCALE*zt);
			LCD_plot(xx,yy,pcolor);
		}
}

#use "6X8L.lib"
#use "12X16L.lib"

typedef struct {
	unsigned long *font;
	unsigned char	lpc;           // lines per character
	unsigned char	Bpcl;          // bytes per character line (1 or 2)
	unsigned char	bpcl;          // bits per character line
} FontInfo;

const static FontInfo fontinfo[]={
{&Font6x8,8,1,6},
{&Font12x16,16,2,12}
};

nodebug void LCD_putchr(unsigned int font,unsigned int row,unsigned int col,char chr ,unsigned long color,long bcolor)
{
int i,j,ii,jj;
unsigned long address;
int data,ddata,aux;

	i=fontinfo[font].lpc;			// lines per character
	ii=fontinfo[font].Bpcl;			// bytes per character line
   jj=fontinfo[font].bpcl;			// bits per character line
	address=*(fontinfo[font].font)+i*ii*(chr-0x20);
	LCD_setwindow (col,col+jj-1,row,row+i-1);
	LCD_WriteCmd(0x2C);
	while(i--){
   	data=xgetint(address);
		aux=(data&0xFF)<<8;			// swap bytes
      data=((data>>8)&0xFF)+aux;
      address+=ii;
      j=jj;
      while(j--){
			if(data&0x8000)
				LCD_WritePixel(color);
      	else {
         	if(bcolor>=0)				// bcolor =-1 => transparent
 					LCD_WritePixel(bcolor);
            else {
 					LCD_WritePixel(0UL);
				}
      	}
         data<<=1;
      }
	}
}

void LCD_printat(unsigned int font,unsigned int row,unsigned int col,char *ptr,unsigned long color,long bcolor)
{
	do {
		LCD_putchr (font,row,col,*ptr++,color,bcolor);
      col+=fontinfo[font].bpcl;
	} 	while (*ptr);
}

nodebug void pattern(void)
{
unsigned int x,y;
unsigned char r=0,g=0,b=0;

	LCD_WriteCmd(0x2C);
  	for(y=0;y<480;y++,g++){
   	if(y>255)
      	b++;
		for(x=0;x<640;x++,r++){
	   	if(x>255)
   	   	b++;
      	LCD_WritePixel((((unsigned long)r)<<16) + (((unsigned long)g)<<8) + b);
      }
   }
}

#ximport "tal.img" tal

/* MAIN PROGRAM */

main()
{
unsigned int i,j,y,oldy;
unsigned char string[5];

	LCD_init();
	while(1){
		LCD_fill(640,480,0xff0000);		// R
		LCD_fill(640,480,0x00ff00);		// G
		LCD_fill(640,480,0x0000ff);		// B
		QUADS();
      pattern();
 		LCD_clear();
		LCD_icon(160,80,320,305,tal);			// show gfx
		LCD_printat(1,420,320,"Tal Wilkenfeld",0x2045C2,0xB33018);
		MsDelay ( 15000 );						// wait 15 secs
 		LCD_clear();
		LCD_rectangle(0,0,639,479,3);
 		for(i=0;i<16;i++) {
			LCD_rectangle(21*i+3,11*i+3,17*i+240,11*i+81,0xff0000);
			LCD_rectangle(13*i+5,13*i+5,19*i+240,13*i+83,0x00ff00);
	   }
		MsDelay ( 1000 );						// wait 1 secs
		LCD_clear();
		LCD_rectangle(10,5,128,45,0xff0000);
		LCD_rectangle(8,3,130,47,0x00ff00);
		LCD_printat(0,20,20,"Cika Electronica",0x0000ff,0x000000);
		MsDelay ( 3000 );						// wait 5 secs
		LCD_clear();
		paraboloide(30,12,20,20,-8.0,8.0,-10.0,10.0,0x0000ff);
		LCD_printat(1,20,20,"Cika Electronica",0x00ff00,0xff0000);
		MsDelay ( 5000 );	 					// wait 5 secs
		elipsoide(45,160.0,30,30,0,6.28,-3.14,3.14,0xff0000);
		LCD_printat(1,50,100,"Cika Electronica",0x00ff00,-1);
		MsDelay ( 5000 );						// wait 5 secs
		LCD_clear();
		LCD_line(0,240,639,240,0x808080);
      for(i=0;i<640;i++){
      	oldy=y;
        	y=(unsigned int)(240-239*sin((2*PI*i)/640)*cos((17*PI*i)/640));
      	if(i==0)
	      	LCD_plot(i,y,0x00ff00);
         else
      		LCD_line(i-1,oldy,i,y,0x00ff00);
      }
      for(i=0;i<640;i++){
      	oldy=y;
      	y=(unsigned int)(240+239*cos((5*PI*i)/640)*cos((PI*i)/640));
      	if(i==0)
	      	LCD_plot(i,y,0xff0000);
         else
      		LCD_line(i-1,oldy,i,y,0xff0000);
      }
      for(i=0;i<640;i++){
      	oldy=y;
      	y=(unsigned int)(240-239*sin(cos(tan(cos((2*PI*i)/640)))));
      	if(i==0)
	      	LCD_plot(i,y,0x0000ff);
         else
      		LCD_line(i-1,oldy,i,y,0x0000ff);
      }
      for(i=0;i<10;i++){
      	LCD_line(64*i,235,64*i,245,0x808080);
      	sprintf(string,"%d",i);
			LCD_printat(0,250,64*i,string,0x808080,-1);
      }
		LCD_printat(1,0,0,"Dermopirifilacion",0xffffff,-1);
		LCD_printat(1,0,500,"Sensor A",0xff0000,-1);
		LCD_printat(1,20,500,"Sensor B",0x00ff00,-1);
		LCD_printat(1,40,500,"Sensor C",0x0000ff,-1);
		MsDelay ( 10000 );						// wait 10 secs
		LCD_setwindow ( 0, 639, 0, 479);
   }
}